/* GREPFV7 - Microsoft C source
Find regular expression in files and archives
=======================================================
Nigel Salt
25 Lower Station Rd
Crayford
Kent
DA1 3PY
UK

Phone +44 322 553260

Email nao@cix.compulink.co.uk
=======================================================
Can process PAK ZIP ARC LZH and ZOO files
   grepfv  expression filepattern[+] tempdir   
   MSoft C V6   
   Argv 1 regular expression
        2 file pattern
        3 Path for temporary files
   FV.COM must be in path along with any necessary
   utilities to unarchive files
   The + activates recursive directory search
*/
/*
   The getfiles function is recursive if search of subdirectories
   is specified.
   This can cause stack overflow which may be remedied by compiling as
   cl -F E000 grepfv.c
   This gives the program a very large stack
*/
#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <dos.h>
#include <string.h>

/* Prototypes */
void dosearch(char *name);
void doarc(char *arcname);
void dofile(char *fname);
void usage(char *error);
char *regstr(char *regexp,char *string);
int regpars(char *r);
void nsfgets(char *buff,int max,FILE *f);
void getfiles(char *p,char *n,int all);

/* Global variables */
typedef struct
{
  char op;
  char *cptr;
} REGSEG;
REGSEG regsegs[64];
int numexp;

char        tempath[256];
char        zippath[256];
char        buff[256];
static char regexp[256];
static char version[]="GREPFV7";
static char copyright[]=" (c) Nigel Salt 1994";
static int  checksum=1422;

void main(int argc,char *argv[])
{
  int all=0;
  int pattend;
  char path[129];
  char name[65];
  char *fslash;
  char *lastslash;
  char *s;
  int count;
  /* Check copyright notice */
  s=copyright;
  count=0;
  while (*s)
     count+=*s++;
  if (count!=checksum)
     {
     fprintf(stderr,"\nCopyright violation");
     exit(1);
     }

  fprintf(stderr,"%s%s\n",version,copyright);
  if (argc<3)
    usage("You must specify an expression and filepattern");

  strcpy(regexp,argv[1]);
  strupr(regexp);
  numexp=regpars(regexp);

  /* Process the file pattern */
  strcpy(path,strupr(argv[2]));
  pattend=strlen(path)-1;
  if (path[pattend]=='+')
    {
    all=1; /* + at end so recurse subdirectories */
    path[pattend]='\0';
    pattend--;
    }

  /* Change forward slashes to backslashes */
  fslash=strchr(path,'/');
  while (fslash!=NULL)
    {
    *fslash='\\';
    fslash=strchr(path,'/');
    }

  /* Split name and path */
  if (path[pattend]=='\\')
    strcpy(name,"*.*");
  else
    {
    lastslash=strrchr(path,'\\');
    if (lastslash==NULL)
      {
      strcpy(name,path);
      path[0]='\0';
      }
    else
      {
      strcpy(name,lastslash+1);
      *(lastslash+1)='\0';
      }
    }

  
  /* Process temporary file path if any */
  if (argc>3)
    {
    strcpy(tempath,argv[3]);
    /* Change forward slashes to backslashes */
    fslash=strchr(tempath,'/');
    while (fslash!=NULL)
      {
      *fslash='\\';
      fslash=strchr(tempath,'/');
      }
    if (tempath[strlen(tempath)-1]!='\\')
      strcat(tempath,"\\");
    }
  else
    tempath[0]='\0';

  /* Now have  
        file pattern in filepatt
        initial path in path
        expression in regexp
        temporary file name in tempath
        all set to 1 if recursion required
  */
  getfiles(path,name,all);            
}

/* Takes Path Name All (ie recurse subdirectories) */
void getfiles(char *p,char *n,int all)
{
  int res;
  unsigned int numfiles=0;
  char cpath[129];
  char npath[129];
  struct find_t cbuff;
  strcpy(cpath,p);
  strcat(cpath,n);
  /* Find normal files */
  res=_dos_findfirst(cpath,0,&cbuff);
  while (res==0)
    {
    if (cbuff.name[0]=='.')
      {
      res=_dos_findnext(&cbuff);
      continue;
      }
    strcpy(npath,p);
    strcat(npath,cbuff.name);
    numfiles++;

    /* DOIT */
    dosearch(npath);
    /* Specify action for each file found */
    res=_dos_findnext(&cbuff);
    }

  /* Search directories if all specified */
  if (all)
    {
    strcpy(cpath,p);
    strcat(cpath,"*.");
    res=_dos_findfirst(cpath,_A_SUBDIR,&cbuff);
    while (res==0)
      {
      if (cbuff.name[0]=='.')
        {
        res=_dos_findnext(&cbuff);
        continue;
        }
      strcpy(npath,p);
      strcat(npath,cbuff.name);
      strcat(npath,"\\");
      getfiles(npath,n,all);
      res=_dos_findnext(&cbuff);
      }
    }
}

void dosearch(char *name)
{
  char *extptr;
  extptr=strrchr(name,'.');
  if (strstr(".ZIP.ARC.LZH.ZOO.PAK.ARJ ",extptr)&&extptr!=NULL)
    doarc(name);
  else
    {
    printf("->SEARCHING %s\n",name);
    dofile(name);
    }
}


/* Gets directory of archive and passes files one at a time to dofile */
void doarc(char *arcname)
{
  FILE *f;
  char *spaceptr;
  char *filelenptr;
  char *slashptr;
  char *extptr;
  int ires;
  static char unzipper[20];
  static char tempfname[256];
  static char temp2fname[256];
  static char fname[256];
  static unsigned long filelen;

  /* Get archive listing */
  printf("==>ARCHIVE %s\n",arcname);
  sprintf(buff,"fv %s>%sgrepfv1.$$$",arcname,tempath);
  ires=system(buff);
  if (ires!=0&&ires!=3)
    {
    printf("\n!!!FV.COM not found or failed on '%s'!!!",arcname);
    return;
    }

  extptr=strrchr(arcname,'.');
  if (strcmp(".ZIP",extptr)==0)
    strcpy(unzipper,"pkunzip -c ");
  else if (strcmp(".PAK",extptr)==0)
    strcpy(unzipper,"pak p ");
  else if (strcmp(".ARC",extptr)==0)
    strcpy(unzipper,"arc p ");
  else if (strcmp(".LZH",extptr)==0)
    strcpy(unzipper,"lharc p ");
  else if (strcmp(".ZOO",extptr)==0)
    strcpy(unzipper,"zoo xp ");
  else if (strcmp(".ARJ",extptr)==0)
    strcpy(unzipper,"arj p ");
  else
    {
    fprintf(stderr,"\nUnrecognised archive '%s'",arcname);
    exit(1);
    }
  strcpy(tempfname,tempath);
  strcpy(temp2fname,tempath);
  strcat(tempfname,"grepfv1.$$$");
  strcat(temp2fname,"grepfv2.$$$");

  /* unzipper now contains the command to unarc to screen
     grepfv1.$$$ contains the list of files
     tempfname contains the full path of temp file 
     temp2fname contains the full path of temp file used to extract
       each file in the archive
  */
  f=fopen(tempfname,"rb");
  if (f==NULL)
    {
    fprintf(stderr,"\nTemporary file '%s' not found",tempfname);
    exit(1);
    }
  fgets(buff,255,f);
  while (!feof(f)&&buff[0]!='=')
    {
    fgets(buff,255,f);
    }

  zippath[0]='\0';
  fgets(buff,255,f);
  /* 
    The format of the output of fv.com is
    filename filesize etc
    directname\
              0 etc
    The output is now in the file pointed to by f
    and the first line is in buff.
  */
  while (!feof(f)&&buff[0]!='=')
    {
    /* Check for path within zip */
    slashptr=strchr(buff,'\\');
    if (slashptr!=NULL)
      {
      *(slashptr+1)='\0';
      strcpy(zippath,buff);
      slashptr=strchr(zippath,'/');
      while (slashptr!=NULL)
        {
        *slashptr='\\';
        slashptr=strchr(zippath,'/');
        }
      fgets(buff,255,f);
      continue;
      }

    /* Check that file length is not zero */
    filelenptr=buff;
    while (!(isdigit(*filelenptr)) && *filelenptr)
      filelenptr++;
    sscanf(filelenptr,"%lu",&filelen);
    if (filelen==0||!*filelenptr)
      {
      fgets(buff,255,f);
      continue;
      }

    /* Check that there is a space terminated file name */
    spaceptr=strchr(buff,' ');
    if (spaceptr!=NULL)
      *spaceptr='\0';
    else
      {
      fgets(buff,255,f);
      continue;
      }

    strcpy(fname,zippath);
    strcat(fname,buff);
    sprintf(buff,"%s %s %s>%s",unzipper,arcname,fname,temp2fname);
    ires=system(buff);
    if (ires)
      printf("\n!!!%s FAILED!!!",buff);
    else
      {
      printf("--->SEARCHING %s\n",fname);
      dofile(temp2fname);
      remove(temp2fname);
      }
    fgets(buff,255,f);
    }
  fclose(f);
  remove(tempfname);
}

/* Search a file */
void dofile(char *name)
{
  FILE *f;
  int i;
  char *cpres=NULL;
  f=fopen(name,"rb");
  if (f==NULL)
    {
    sprintf(buff,"Cannot open '%s'",name);
    usage(buff);
    }
  nsfgets(buff,81,f);
  while (!feof(f))
    {
    for (i=0;i<numexp;i++)
      {
      cpres=regstr(regsegs[i].cptr,buff);
      if (cpres!=NULL&&regsegs[i].op=='|')
        break;
      if (cpres==NULL&&regsegs[i].op!='|')
        break;
      }
    if (cpres!=(char *)NULL)
      printf("%s\n",buff);
    nsfgets(buff,81,f);
    }
  fclose(f);
}

void nsfgets(char *buff,int max,FILE *f)
{
  int i;
  int inch;
  inch=0;
  *buff='\0';
  while (inch<' '&&!feof(f))
    inch=getc(f);
  /* Now have  a printable character in inch or reached EOF*/
  if (feof(f))
    return;
  for (i=0;i<(max-1)&&inch>=' '&&!feof(f);i++)
    {
    *(buff+i)=(char)inch;
    inch=getc(f);
    }
  if (i==(max-1))
    ungetc(inch,f);
  *(buff+i)='\0';
  return;
}

void usage(char *error)
{
  fprintf(stderr,"\n%s",error);
  fprintf(stderr,"\nUsage:");
  fprintf(stderr,"\n\tgrepfv \"regexp\" filepattern[+] tempdir");
  fprintf(stderr,"\nREGULAR EXPRESSIONS: ");
  fprintf(stderr,"\n\\    treat next char as literal");
  fprintf(stderr,"\n.     match any single character");
  fprintf(stderr,"\n*     match 0 or more characters");
  fprintf(stderr,"\n|     preceeding expression OR following");
  fprintf(stderr,"\n&     preceeding expression AND following");
  fprintf(stderr,"\n");
  fprintf(stderr,"\nThe optional + on the end of filepattern will cause grepfv to"); 
  fprintf(stderr,"\nsearch subdirectories too");
  exit(1);
}


/* Looks for AND & and OR | chars and break up regexp into array
   of structures regsegs returns number of segs found */
int regpars(char *r)
{
  int regind=0;
  char *rpos=r;
  regsegs[0].cptr=r;
  while (*r)
    {
    switch(*r)
      {
      case '\\':
        r++;
        break;
      case '|':
      case '&':
        regsegs[regind].op=*r;
        *r='\0';
        regind++;
        if (regind>63)
          usage("YOU HAVE MORE THAN 64 SUBEXPRESSIONS IN YOUR REGULAR EXPRESSION");
        regsegs[regind].cptr=r+1;
        break;
      default:
        break;
      }
    r++;
    }
  return regind+1;
}

/* Do the search using wild cards */
char *regstr(char *r,char *s)
{
  int res;
  char *spos,*rpos;

  res=0;
  while (*s&&res!=1)
    {
    res=1;
    spos=s;
    rpos=r;
    while (*rpos&&res==1)
      {
      switch (*rpos)
        {
        case '\\':
          rpos++;
          if (toupper(*spos)!=*rpos)
            res=0;
          break;
        case '.':
          break;
        case '*':
          rpos++;
          while (toupper(*spos)!=*rpos&&*spos)
            spos++;
        default:
          if (toupper(*spos)!=*rpos)
            res=0;
        }
      rpos++;
      spos++;
      }
    s++;
    }
  return (res==1?s:(char *)NULL);
}
